@@ -2,44 +2,40 @@
 Defines GenericForeignFileField, a subclass of GenericRelation from
 django.contrib.contenttypes.
 """
+import itertools
+import operator
+
 import six
+from six.moves import reduce
 
-import inspect
-from operator import attrgetter
-
+import django
 from django.contrib.contenttypes.models import ContentType
-from django.contrib.contenttypes.generic import GenericInlineModelAdmin
 from django.core.exceptions import ObjectDoesNotExist
 from django.core.files.base import File
 from django.core.files.uploadedfile import UploadedFile
 from django.db import connection, router, models
-from django.db.models.deletion import DO_NOTHING
 from django.db.models.fields.files import FieldFile, FileDescriptor
-from django.db.models.fields.related import RelatedField, Field, ManyToOneRel
-from django.contrib.contenttypes.generic import GenericRelation
-from django.utils.functional import curry
 
-try:
-    from django.contrib.contenttypes.fields import GenericRel
-except ImportError:
-    from django.contrib.contenttypes.generic import GenericRel
-
-try:
-    from django.db.models.fields.related import ForeignObject, ForeignObjectRel
-except ImportError:
-    # Django <= 1.5
-    class ForeignObject(RelatedField, Field):
-        pass
-
-    class ForeignObjectRel(ManyToOneRel):
-        def __init__(self, field, to, **kwargs):
-            self.related_query_name = kwargs.pop('related_query_name', None)
-            self.field = field
-            super(ForeignObjectRel, self).__init__(to, self.field.name, **kwargs)
+from django.contrib.contenttypes.admin import GenericInlineModelAdmin
+from django.contrib.contenttypes.fields import GenericRelation, GenericRel
 
+from generic_plus.compat import compat_rel, compat_rel_to
 from generic_plus.forms import (
     generic_fk_file_formfield_factory, generic_fk_file_widget_factory)
 
+try:
+    from django.utils.functional import curry
+except ImportError:
+    # You can't trivially replace this with `functools.partial` because this binds
+    # to classes and returns bound instances, whereas functools.partial (on
+    # CPython) is a type and its instances don't bind.
+    def curry(_curried_func, *args, **kwargs):
+        def _curried(*moreargs, **morekwargs):
+            merged = dict(**kwargs)
+            merged.update(morekwargs)
+            return _curried_func(*(args + moreargs), **merged)
+        return _curried
+
 
 class GenericForeignFileField(GenericRelation):
     """
@@ -75,8 +71,10 @@ class GenericForeignFileField(GenericRelation):
     file_descriptor_cls = FileDescriptor
     file_field_cls = models.FileField
     rel_file_field_name = 'file'
+    field_identifier_field_name = None
 
-    def __init__(self, to, rel_file_field_name=None, **kwargs):
+    def __init__(self, to, rel_file_field_name=None, field_identifier="",
+            missing_file_fallback=True, **kwargs):
         """
         Parameters
         ----------
@@ -84,11 +82,21 @@ class GenericForeignFileField(GenericRelation):
         rel_file_field_name : str
             Name of the FileField on the generic related model (e.g. "image",
             "file" [the default])
+        field_identifier : str
+            A string to uniquely identify the field on the model, allowing
+            multiple GenericForeignFileFields to point to a single model class
+        missing_file_fallback : bool
+            If set to True (the default), the GenericForeignFileField widget
+            will show the admin's file field widget in the event that there is
+            a value on the model for the file field, but no corresponding row
+            in the table with the generic foreign key.
         """
         self.rel_file_field_name = rel_file_field_name or self.rel_file_field_name
+        self.field_identifier = field_identifier
+        self.missing_file_fallback = missing_file_fallback
 
         self.file_kwargs = {
-            'editable': False,
+            'editable': (django.VERSION > (1, 10)),
             'default': '',
             'blank': True,
             'upload_to': kwargs.pop('upload_to', None),
@@ -102,24 +110,14 @@ class GenericForeignFileField(GenericRelation):
             'height_field': kwargs.pop('height_field', None),
         }
 
-        symmetrical = kwargs.pop('symmetrical', True)
-
-        if issubclass(GenericRel, ForeignObjectRel):
-            # Django 1.6
-            kwargs['rel'] = GenericRel(self, to,
-                related_name=kwargs.pop('related_name', None),
-                limit_choices_to=kwargs.pop('limit_choices_to', None))
-            kwargs['rel'].on_delete = DO_NOTHING
-        else:
-            # Django <= 1.5
-            kwargs['rel'] = GenericRel(to,
-                                related_name=kwargs.pop('related_name', None),
-                                limit_choices_to=kwargs.pop('limit_choices_to', None),
-                                symmetrical=symmetrical)
+        kwargs['rel'] = GenericRel(self, to,
+            related_name=kwargs.pop('related_name', None),
+            limit_choices_to=kwargs.pop('limit_choices_to', None))
 
         # Override content-type/object-id field names on the related class
         self.object_id_field_name = kwargs.pop("object_id_field", "object_id")
         self.content_type_field_name = kwargs.pop("content_type_field", "content_type")
+        self.field_identifier_field_name = kwargs.pop("field_identifier_field", self.field_identifier_field_name)
 
         self.for_concrete_model = kwargs.pop("for_concrete_model", True)
 
@@ -130,33 +128,45 @@ class GenericForeignFileField(GenericRelation):
             'max_length': self.file_kwargs['max_length'],
         })
 
-        if isinstance(self, ForeignObject):
-            # Django 1.6
-            super(GenericRelation, self).__init__(
-                to, to_fields=[],
-                from_fields=[self.object_id_field_name], **kwargs)
-        else:
-            # Django <= 1.5
-            models.Field.__init__(self, **kwargs)
+        if django.VERSION > (1, 9):
+            kwargs['on_delete'] = models.CASCADE
+
+        super(GenericRelation, self).__init__(to,
+            from_fields=[self.object_id_field_name], to_fields=[], **kwargs)
 
         self.file_kwargs['db_column'] = kwargs.get('db_column', self.name)
 
+    def get_cached_value(self, instance, **kwargs):
+        cache_name = self.get_cache_name()
+        if django.VERSION > (2, 0):
+            return super(GenericForeignFileField, self).get_cached_value(
+                instance, **kwargs)
+        else:
+            return instance.__dict__[cache_name]
+
+    def set_cached_value(self, instance, value):
+        cache_name = self.get_cache_name()
+        if django.VERSION > (2, 0):
+            super(GenericForeignFileField, self).set_cached_value(instance, value)
+        else:
+            instance.__dict__[cache_name] = value
+
     def contribute_to_class(self, cls, name):
         self.generic_rel_name = '%s_generic_rel' % name
         self.raw_file_field_name = '%s_raw' % name
         self.file_field_name = name
 
+        self.set_attributes_from_name(name)
+        self.file_kwargs['db_column'] = self.db_column or self.attname
+
         # Save a reference to which model this class is on for future use
         self.model = cls
 
-        contribute_to_class_args = inspect.getargspec(
-                models.Field.contribute_to_class).args
+        super(GenericRelation, self).contribute_to_class(cls, name, **{
+            ('private_only' if django.VERSION > (1, 10) else 'virtual_only'): True,
+        })
 
-        if 'virtual_only' in contribute_to_class_args:
-            # Django 1.6+
-            cls._meta.add_virtual_field(self)
-
-        super(GenericRelation, self).contribute_to_class(cls, name)
+        self.column = self.file_kwargs['db_column']
 
         if not isinstance(self.file_field_cls, models.ImageField):
             self.file_kwargs.pop('width_field', None)
@@ -167,22 +177,32 @@ class GenericForeignFileField(GenericRelation):
             if not self.file_kwargs['height_field']:
                 del self.file_kwargs['height_field']
 
-        self.__dict__['file_field'] = self.file_field_cls(**self.file_kwargs)
+        self.__dict__['file_field'] = self.file_field_cls(name=name, **self.file_kwargs)
         ### HACK: manually fix creation counter
         self.file_field.creation_counter = self.creation_counter
 
         # This calls contribute_to_class() for the FileField
-        cls.add_to_class(self.file_field_name, self.file_field)
+        parents = cls._meta.parents.keys()
+        parent_field_names = []
+        if parents:
+            parent_fields = reduce(operator.add, [p._meta.local_fields for p in parents], [])
+            parent_field_names = [f.name for f in parent_fields]
+        # Don't duplicate the field when inherited from a parent model
+        if self.file_field_name not in parent_field_names:
+            # Don't add field to proxy models
+            if not cls._meta.proxy:
+                cls.add_to_class(self.file_field_name, self.file_field)
 
         # Add the descriptor for the generic relation
-        generic_descriptor = GenericForeignFileDescriptor(self, self.file_field)
+        generic_descriptor = GenericForeignFileDescriptor(self, self.file_field,
+            for_concrete_model=self.for_concrete_model)
         # We use self.__dict__ to avoid triggering __get__()
         self.__dict__['generic_descriptor'] = generic_descriptor
         setattr(cls, self.generic_rel_name, generic_descriptor)
 
         # Add the descriptor for the FileField
         file_descriptor = GenericForeignFileDescriptor(self, self.file_field,
-            is_file_field=True)
+            is_file_field=True, for_concrete_model=self.for_concrete_model)
         self.__dict__['file_descriptor'] = file_descriptor
         setattr(cls, self.file_field_name, file_descriptor)
 
@@ -195,79 +215,55 @@ class GenericForeignFileField(GenericRelation):
         setattr(cls, self.raw_file_field_name, self.file_descriptor_cls(self.file_field))
 
     def is_cached(self, instance):
-        return hasattr(instance, self.get_cache_name())
+        if django.VERSION > (2, 0):
+            return super(GenericForeignFileField, self).is_cached(instance)
+        else:
+            return hasattr(instance, self.get_cache_name())
+
+    def get_prefetch_queryset(self, instances, queryset=None):
+        models = set([type(i) for i in instances])
+
+        # Handle case where instances are different models (and consequently,
+        # different content types)
+        if len(models) > 1:
+            bulk_qsets = []
+            for model, group in itertools.groupby(instances, type):
+                model_instances = list(group)
+                field = getattr(model, self.name)
+                bulk_qsets.append(field.bulk_related_objects(model_instances))
+            bulk_qset = reduce(operator.or_, bulk_qsets)
+
+            def rel_obj_attr(rel_obj):
+                content_type = getattr(rel_obj, "%s_id" % self.content_type_field_name)
+                object_id = getattr(rel_obj, self.object_id_field_name)
+                return (content_type, object_id)
+
+            def get_ctype_obj_id(obj):
+                field = getattr(obj.__class__, self.name)
+                content_type = ContentType.objects.get_for_model(obj, field.for_concrete_model)
+                return (content_type.pk, obj._get_pk_val())
+
+            return (bulk_qset,
+                rel_obj_attr,
+                get_ctype_obj_id,
+                True,
+                self.attname) + (() if django.VERSION < (2, 0) else (True,))
 
-    def get_prefetch_query_set(self, instances):
         return (self.bulk_related_objects(instances),
-            attrgetter(self.object_id_field_name),
+            operator.attrgetter(self.object_id_field_name),
             lambda obj: obj._get_pk_val(),
             True,
-            self.attname)
+            self.attname) + (() if django.VERSION < (2, 0) else (True,))
 
-    def south_init(self):
+    def bulk_related_objects(self, *args, **kwargs):
         """
-        This method is called by south before it introspects the field.
+        Return all objects related to ``objs`` via this ``GenericRelation``.
 
-        South assumes that this is a related field if self.rel is set and it
-        is not None. While this is a reasonable assumption, and it is *mostly*
-        true for GenericForeignFileField, it is incorrect as far as South is
-        concerned; we need South to treat this as a FileField so that
-        it creates a column in the containing model.
-
-        To deal with this situation we conditionally return the same values as
-        FileField from get_internal_type() and db_type() while south is
-        introspecting the field, and otherwise return the values that would be
-        returned by a GenericRelation (which are the same as those returned
-        by a ManyToManyField)
-
-        self.south_executing is the basis for the conditional logic. It is set
-        to True in this method (south_init()) and then back to False in
-        GenericForeignFileField.post_create_sql().
         """
-        self.south_executing = True
-        self._rel = self.rel
-        self.rel = None
-
-    def post_create_sql(self, style, db_table):
-        """
-        This method is called after south is done introspecting the field.
-
-        See GenericForeignFileField.south_init() for more documentation
-        about the reason this is overridden here.
-        """
-        self.south_executing = False
-        if self.rel is None and hasattr(self, '_rel'):
-            self.rel = self._rel
-        return []
-
-    def get_internal_type(self):
-        """
-        Related to the implementation of db_type(), returns the pre-existing
-        Django Field class whose database column is the same as the current
-        field class, if such a class exists.
-
-        See GenericForeignFileField.south_init() for more documentation
-        about the reason this is overridden here.
-        """
-        if self.south_executing:
-            return 'FileField'
-        else:
-            # super() returns 'ManyToManyField'
-            return super(GenericForeignFileField, self).get_internal_type()
-
-    def db_type(self, connection):
-        """
-        Returns the database column data type for this field, for the provided
-        connection.
-
-        See GenericForeignFileField.south_init() for more documentation
-        about the reason this is overridden here.
-        """
-        if self.south_executing:
-            return models.Field.db_type(self, connection)
-        else:
-            # super() returns None
-            return super(GenericForeignFileField, self).db_type(connection)
+        qs = super(GenericForeignFileField, self).bulk_related_objects(*args, **kwargs)
+        if self.field_identifier_field_name:
+            qs = qs.filter(**{"%s__exact" % self.field_identifier_field_name: self.field_identifier})
+        return qs
 
     def save_form_data(self, instance, data):
         super(GenericForeignFileField, self).save_form_data(instance, data)
@@ -297,9 +293,7 @@ class GenericForeignFileField(GenericRelation):
             instance.save()
 
     def formfield(self, **kwargs):
-        factory_kwargs = {
-            'related': self.related,
-        }
+        factory_kwargs = {'related': compat_rel(self)}
         widget = kwargs.pop('widget', None) or generic_fk_file_widget_factory(**factory_kwargs)
         formfield = kwargs.pop('form_class', None) or generic_fk_file_formfield_factory(widget=widget, **factory_kwargs)
         widget.parent_admin = formfield.parent_admin = kwargs.pop('parent_admin', None)
@@ -316,13 +310,13 @@ class GenericForeignFileField(GenericRelation):
         })
         return super(GenericForeignFileField, self).formfield(**kwargs)
 
-    def get_inline_admin_formset(self, formset_cls=None, **kwargs):
+    def get_inline_admin_formset(self, formset_cls=None, form_attrs=None, **kwargs):
         from generic_plus.forms import generic_fk_file_formset_factory, BaseGenericFileInlineFormSet
 
         formset_cls = formset_cls or BaseGenericFileInlineFormSet
 
         attrs = {
-            'model': self.rel.to,
+            'model': compat_rel_to(self),
             'default_prefix': self.name,
             'field': self,
             'formset_kwargs': kwargs.pop('formset_kwargs', None) or {},
@@ -356,7 +350,8 @@ class GenericForeignFileField(GenericRelation):
                         formset_attrs=kwargs,
                         field=self.field,
                         prefix=self.default_prefix,
-                        formfield_callback=curry(self.formfield_for_dbfield, request=request))
+                        formfield_callback=curry(self.formfield_for_dbfield, request=request),
+                        form_attrs=form_attrs)
                     if getattr(self, 'default_prefix', None):
                         formset.default_prefix = self.default_prefix
                     return formset
@@ -365,6 +360,19 @@ class GenericForeignFileField(GenericRelation):
 
         return type('GenericForeignFileInline', (GenericForeignFileInline,), attrs)
 
+    @property
+    def get_path_info(self):
+        raise AttributeError("'%s' object has no attribute 'get_path_info'" % type(self).__name__)
+
+    @property
+    def get_lookup_constraint(self):
+        raise AttributeError("'%s' object has no attribute 'get_lookup_constraint'" % type(self).__name__)
+
+    def get_attname_column(self):
+        attname = self.get_attname()
+        column = self.db_column or attname
+        return attname, column
+
 
 class GenericForeignFileDescriptor(object):
 
@@ -378,7 +386,6 @@ class GenericForeignFileDescriptor(object):
         if instance is None:
             return self.field
 
-        cache_name = self.field.get_cache_name()
         file_val = None
 
         if self.is_file_field:
@@ -386,41 +393,20 @@ class GenericForeignFileDescriptor(object):
 
         # Dynamically create a class that subclasses the related model's
         # default manager.
-        rel_model = self.field.rel.to
+        rel_model = compat_rel_to(self.field)
         superclass = rel_model._default_manager.__class__
         RelatedManager = create_generic_related_manager(superclass)
 
         qn = connection.ops.quote_name
 
-        if hasattr(instance.__class__._default_manager, 'prefetch_related'):
-            # Django 1.4+
-            manager_kwargs = {
-                'prefetch_cache_name': self.field.attname,
-            }
-        else:
-            # Django <= 1.3
-            manager_kwargs = {
-                'join_table': qn(self.field.m2m_db_table()),
-            }
+        manager_kwargs = {
+            'prefetch_cache_name': self.field.attname,
+        }
 
-        if hasattr(self.field.rel, 'symmetrical'):
-            # Django <= 1.5
-            manager_kwargs['symmetrical'] = (self.field.rel.symmetrical and instance.__class__ == rel_model)
-
-        if hasattr(self.field, 'get_joining_columns'):
-            join_cols = self.field.get_joining_columns(reverse_join=True)[0]
-        else:
-            join_cols = [self.field.m2m_column_name(), self.field.m2m_reverse_name()]
+        join_cols = self.field.get_joining_columns(reverse_join=True)[0]
 
         ct_manager = ContentType.objects.db_manager(instance._state.db)
-        try:
-            content_type = ct_manager.get_for_model(instance, for_concrete_model=self.for_concrete_model)
-        except TypeError:
-            # Django <= 1.5
-            if not self.for_concrete_model:
-                raise
-            else:
-                content_type = ct_manager.get_for_model(instance)
+        content_type = ct_manager.get_for_model(instance, for_concrete_model=self.for_concrete_model)
 
         manager = RelatedManager(
             model=rel_model,
@@ -431,6 +417,7 @@ class GenericForeignFileDescriptor(object):
             content_type=content_type,
             content_type_field_name=self.field.content_type_field_name,
             object_id_field_name=self.field.object_id_field_name,
+            field_identifier_field_name=self.field.field_identifier_field_name,
             **manager_kwargs)
 
         if not manager.pk_val:
@@ -440,33 +427,18 @@ class GenericForeignFileDescriptor(object):
                 return manager
 
             try:
-                val = getattr(instance, cache_name)
-            except AttributeError:
+                val = self.field.get_cached_value(instance)
+            except KeyError:
                 db = manager._db or router.db_for_read(rel_model, instance=instance)
-                query = manager.core_filters or {
-                    '%s__pk' % manager.content_type_field_name: manager.content_type.id,
-                    '%s__exact' % manager.object_id_field_name: manager.pk_val,
-                }
+                qset = superclass.get_queryset(manager).using(db)
 
                 try:
-                    prev_object_id_field = rel_model._meta.get_field(
-                        'prev_%s' % manager.object_id_field_name)
-                except models.FieldDoesNotExist:
-                    pass
-                else:
-                    query['%s__isnull' % prev_object_id_field.attname] = True
-
-                qset = superclass.get_query_set(manager).using(db)
-
-                try:
-                    val = qset.get(**query)
+                    val = qset.get(**manager.core_filters)
                 except rel_model.DoesNotExist:
-                    if not self.is_file_field:
-                        return None
                     val = None
 
             self.set_file_value(instance, file_val, obj=val)
-            setattr(instance, self.field.get_cache_name(), val)
+            self.field.set_cached_value(instance, val)
         return instance.__dict__[self.file_field.name]
 
     def set_file_value(self, instance, value, obj=None):
@@ -494,9 +466,9 @@ class GenericForeignFileDescriptor(object):
             instance.__dict__[self.file_field.name] = value
             return
 
-        if isinstance(obj, self.field.rel.to):
+        if isinstance(obj, compat_rel_to(self.field)):
             value = getattr(obj, self.field.rel_file_field_name)
-        elif isinstance(value, self.field.rel.to):
+        elif isinstance(value, compat_rel_to(self.field)):
             obj = value
             value = getattr(obj, self.field.rel_file_field_name)
 
@@ -526,8 +498,8 @@ class GenericForeignFileDescriptor(object):
         if instance is None:
             raise AttributeError("Manager must be accessed via instance")
 
-        if isinstance(value, self.field.rel.to) or value is None:
-            setattr(instance, self.field.get_cache_name(), value)
+        if isinstance(value, compat_rel_to(self.field)) or value is None:
+            self.field.set_cached_value(instance, value)
 
         if self.is_file_field:
             self.set_file_value(instance, value)
@@ -536,7 +508,7 @@ class GenericForeignFileDescriptor(object):
             manager.clear()
             if value is None:
                 return
-            if isinstance(value, self.field.rel.to):
+            if isinstance(value, compat_rel_to(self.field)):
                 rel_obj_file = getattr(value, self.field.rel_file_field_name)
                 file_val = rel_obj_file.path if rel_obj_file else None
                 setattr(instance, self.field.file_field_name, file_val)
@@ -547,7 +519,7 @@ class GenericForeignFileDescriptor(object):
                     file_val = field_value.path if field_value else None
                     setattr(instance, self.field.file_field_name, file_val)
                     manager.add(obj)
-                    setattr(instance, self.field.get_cache_name(), value)
+                self.field.set_cached_value(instance, value)
 
 
 def create_generic_related_manager(superclass):
@@ -560,60 +532,54 @@ def create_generic_related_manager(superclass):
 
         def __init__(self, model=None, instance=None, symmetrical=None,
                      source_col_name=None, target_col_name=None, content_type=None,
-                     content_type_field_name=None, object_id_field_name=None, **kwargs):
+                     content_type_field_name=None, object_id_field_name=None,
+                     prefetch_cache_name=None,
+                     field_identifier_field_name=None, **kwargs):
             super(GenericRelatedObjectManager, self).__init__()
             self.model = model
             self.content_type = content_type
             self.symmetrical = symmetrical
             self.instance = instance
-            if 'join_table' in kwargs:
-                # django <= 1.3
-                self.join_table = kwargs.pop('join_table')
-                self.join_table = model._meta.db_table
-                self.core_filters = kwargs.pop('core_filters', None) or {}
-            else:
-                # django 1.4+
-                self.core_filters = {
-                    '%s__pk' % content_type_field_name: content_type.id,
-                    '%s__exact' % object_id_field_name: instance._get_pk_val(),
-                }
-            if 'prefetch_cache_name' in kwargs:
-                # django 1.4+
-                self.prefetch_cache_name = kwargs['prefetch_cache_name']
+            self._field = kwargs.pop('field', None)
+            self.file_field_name = self._field.file_field_name
+            self.core_filters = {
+                '%s__pk' % content_type_field_name: content_type.id,
+                '%s__exact' % object_id_field_name: instance._get_pk_val(),
+            }
+            if field_identifier_field_name:
+                self.core_filters['%s__exact' % field_identifier_field_name] = getattr(self._field, field_identifier_field_name)
+
+            self.prefetch_cache_name = prefetch_cache_name
             self.source_col_name = source_col_name
             self.target_col_name = target_col_name
             self.content_type_field_name = content_type_field_name
             self.object_id_field_name = object_id_field_name
             self.pk_val = self.instance._get_pk_val()
-            # Change from django.contrib.contenttypes.generic.create_generic_related_manager()
-            self._field = kwargs.pop('field', None)
-            self.file_field_name = self._field.file_field_name
 
-        def get_query_set(self):
-            if hasattr(self, 'prefetch_cache_name'):
-                try:
-                    return self.instance._prefetched_objects_cache[self.prefetch_cache_name]
-                except (AttributeError, KeyError):
-                    pass
+        def get_queryset(self):
+            try:
+                return self.instance._prefetched_objects_cache[self.prefetch_cache_name]
+            except (AttributeError, KeyError):
+                pass
             db = self._db or router.db_for_read(self.model, instance=self.instance)
             query = {
                 ('%s__pk' % self.content_type_field_name): self.content_type.id,
                 ('%s__exact' % self.object_id_field_name): self.pk_val,
             }
-            return superclass.get_query_set(self).using(db).filter(**query)
+            return superclass.get_queryset(self).using(db).filter(**query)
 
-        def get_prefetch_query_set(self, instances):
+        def get_prefetch_queryset(self, instances, queryset=None):
             db = self._db or router.db_for_read(self.model, instance=instances[0])
             query = {
                 ('%s__pk' % self.content_type_field_name): self.content_type.id,
                 ('%s__in' % self.object_id_field_name): set(obj._get_pk_val() for obj in instances),
             }
-            qs = super(GenericRelatedObjectManager, self).get_query_set().using(db).filter(**query)
-            return (qs,
-                    attrgetter(self.object_id_field_name),
+            qs = super(GenericRelatedObjectManager, self).get_queryset()
+            return (qs.using(db).filter(**query),
+                    operator.attrgetter(self.object_id_field_name),
                     lambda obj: obj._get_pk_val(),
                     False,
-                    self.prefetch_cache_name)
+                    self.prefetch_cache_name) + (() if django.VERSION < (2, 0) else (False,))
 
         def add(self, *objs):
             for obj in objs:
@@ -669,23 +635,3 @@ def create_generic_related_manager(superclass):
         create.alters_data = True
 
     return GenericRelatedObjectManager
-
-
-try:
-    from south.modelsinspector import add_introspection_rules
-except ImportError:
-    pass
-else:
-    add_introspection_rules(rules=[
-        (
-            (GenericForeignFileField,),
-            [],
-            {
-                "to": ["rel.to", {}],
-                "symmetrical": ["rel.symmetrical", {"default": True}],
-                "object_id_field": ["object_id_field_name", {"default": "object_id"}],
-                "content_type_field": ["content_type_field_name", {"default": "content_type"}],
-                "blank": ["blank", {"default": True}],
-            },
-        ),
-    ], patterns=["^generic_plus\.fields\.GenericForeignFileField"])
