@@ -1,8 +1,8 @@
-import six
-
 import re
 
-from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
+import six
+import django
+from django.core.exceptions import ObjectDoesNotExist
 from django.forms.widgets import Input
 from django.conf import settings
 from django.contrib.admin import helpers
@@ -10,6 +10,8 @@ from django.contrib.admin.sites import site
 from django.db.models.fields.files import FieldFile
 from django.template.loader import render_to_string
 
+from generic_plus.compat import compat_rel_to
+
 
 class GenericForeignFileWidget(Input):
 
@@ -20,13 +22,21 @@ class GenericForeignFileWidget(Input):
         self.attrs = attrs.copy() if attrs is not None else {}
 
     def get_context_data(self, name, value, attrs=None, bound_field=None):
-        final_attrs = self.build_attrs(attrs, type=self.input_type, name=name)
+        attrs = attrs or {}
+        attrs.update({
+            'type': self.input_type,
+            'name': name,
+        })
+        final_attrs = self.build_attrs(attrs)
 
         formfield = getattr(bound_field, 'field', None)
         related = getattr(formfield, 'related', None)
         dbfield = getattr(related, 'field', None)
         file_field = getattr(dbfield, 'file_field', None)
-        rel_model = getattr(getattr(dbfield, 'rel', None), 'to', None)
+        if django.VERSION > (1, 9):
+            rel_model = getattr(getattr(dbfield, 'remote_field', None), 'model', None)
+        else:
+            rel_model = getattr(getattr(dbfield, 'rel', None), 'to', None)
 
         obj = None
         file_value = ''
@@ -40,19 +50,19 @@ class GenericForeignFileWidget(Input):
                 file_value = getattr(obj, self.field.name).name
             value = getattr(obj, 'pk', None)
         elif rel_model and isinstance(value, six.string_types) and not value.isdigit():
-            try:
-                obj = rel_model.objects.get(**{self.field.rel_file_field_name: value})
-            except ObjectDoesNotExist:
-                obj = None
-                file_value = value
-                value = None
-            except MultipleObjectsReturned:
-                file_value = value
-                if bound_field and getattr(bound_field, 'form', None):
-                    value = bound_field.form.data.get("%s-0-id" % bound_field.form.add_prefix(name))
-            else:
-                file_value = value
-                value = obj.pk
+            file_value = value
+            if bound_field and getattr(bound_field, 'form', None):
+                if not bound_field.form.prefix or name.startswith(bound_field.form.prefix):
+                    formset_prefix = name
+                else:
+                    formset_prefix = bound_field.form.add_prefix(name)
+                generic_field = getattr(bound_field.db_file_field, 'generic_field', None)
+                if generic_field:
+                    pk_name = compat_rel_to(generic_field)._meta.pk.name
+                else:
+                    pk_name = 'id'
+                value = bound_field.form.data.get("%s-0-%s" % (
+                    formset_prefix, pk_name))
         if rel_model and isinstance(value, six.integer_types) or (isinstance(value, six.string_types) and value.isdigit()):
             if not obj or not file_value:
                 try:
@@ -80,7 +90,6 @@ class GenericForeignFileWidget(Input):
         final_attrs['value'] = value or ''
 
         formset = self.get_inline_admin_formset(name, value, instance=obj, bound_field=bound_field)
-
         return {
             'instance': obj,
             'value': value,
@@ -121,8 +130,12 @@ class GenericForeignFileWidget(Input):
             'prefix': name,
         }
         if instance:
-            formset_kwargs['instance'] = instance.content_object
-            formset_kwargs['queryset'] = instance.__class__._default_manager
+            formset_kwargs['instance'] = bound_field.form.instance
+            qs = instance.__class__._default_manager
+            if self.field.field_identifier_field_name:
+                field_identifier = getattr(instance, self.field.field_identifier_field_name)
+                qs = qs.filter(field_identifier=field_identifier)
+            formset_kwargs['queryset'] = qs
 
         formset = FormSet(**formset_kwargs)
         parent_admin = getattr(self, 'parent_admin', None)
@@ -147,7 +160,7 @@ class GenericForeignFileWidget(Input):
         return data.get(name, files.get(name, None))
 
 
-def generic_fk_file_widget_factory(widget_cls, related=None, **attrs):
+def generic_fk_file_widget_factory(widget_cls=GenericForeignFileWidget, related=None, **attrs):
     return type('GenericForeignFileWidget', (widget_cls,), attrs)
     widget_attrs = {
         '__module__': widget_cls.__module__,
